### PersefoneDesktop - a GUI to the Persefone model of agriculture and ecosystems
###
### This file links the QML UI designs to the underlying model.
###

global model = nothing
global landcovermap = nothing
global runsimulation = nothing
global timer = nothing
global launching = false

const configfile = Observable("guiparams.toml")
const running = Observable(false)
const ticks = Observable(0)
const date = Observable(today())
const progress = Observable(0.0)
const delay = Observable(0.5)
const runbuttontext = Observable(">>")
const runbuttontip = Observable("Run")
const mapimage = Observable(Figure().scene)

global guiproperties =
    ["running" => running,
     "ticks" => ticks,
     "date" => date,
     "delay" => delay,
     "progress" => progress,
     "runbuttontext" => runbuttontext,
     "runbuttontip" => runbuttontip]

function newsimulation()
    global model, landcovermap
    running[] = false
    progress[] = 0.0
    model = initialise(configfile[])
    landcovermap = rotr90(load(@param(world.landcovermap)))
    date[] = model.date
    println("Model initialised.")
end

function loadsimulation(filename)
    global model
    fn = convert(filename, String)
    startswith(fn, "file://") && (fn = fn[8:end])
    model = loadmodelobject(fn)
    running[] = false
    landcovermap = rotr90(load(@param(world.landcovermap)))
    date[] = model.date
    progress[] = (date[]-@param(core.startdate)) /
        (@param(core.enddate)-@param(core.startdate))
    @emit updateMakie()
end

function savesimulation(filename)
    global model
    fn = convert(filename, String)
    startswith(fn, "file://") && (fn = fn[8:end])
    savemodelobject(model, fn)
end

function Base.convert(s::QString, ::Type{<:AbstractString})
    s2 = ""
    for l in s
        s2 *= l
    end
    s2
end

#TODO saveoutput()

function nextstep()
    global model
    if date[] >= @param(core.enddate)
        running[] = false
        runbuttontext[] = ">>"
        runbuttontip[] = "Run"
        return
    elseif date[] < model.date
        # If we've "scrolled back" in time, step forward again without
        # having to resimulate (see `previousstep()`)
        date[] += Day(1)
    else
        # If we're already displaying the newest model update, we need to
        # keep simulating
        stepsimulation!(model)
        date[] = model.date
    end
    progress[] = (date[]-@param(core.startdate)) /
        (@param(core.enddate)-@param(core.startdate))
    @emit updateMakie()
    println("Updated model: $(date[]-Day(1))")
end

function previousstep()
    # Since we store all model output in a dataframe, we can simply
    # "scroll back" in time and visualise the state of the model at
    # a previous date
    global model
    date[] <= @param(core.startdate) && return
    date[] -= Day(1)
    progress[] = (date[]-@param(core.startdate)) /
        (@param(core.enddate)-@param(core.startdate))
    @emit updateMakie()
    println("Backtracked model: $(date[]-Day(1))")
end

#XXX This could also be done with Channels rather than ResumableFunctions,
# which would save us one dependency (Channels are inbuilt). However, the
# ResumableFunctions version seemed to work more smoothly.
@resumable function createrunfunction()
    i = round(delay[]*10)
    while running[]
        i -= 1
        if i <= 0
            nextstep()
            i = round(delay[]*10)
        end
        # to reduce interface lag, always only sleep for 0.1s
        sleep(0.1)
        @yield 1
    end
end

function togglerunning()
    global runsimulation, timer
    if running[]
        running[] = false
        QML.stop(timer)
        runbuttontext[] = ">>"
        runbuttontip[] = "Run"
    else
        running[] = true
        QML.start(timer)
        runbuttontext[] = "||"
        runbuttontip[] = "Pause"
        runsimulation = createrunfunction()
    end
end
 
function render_map(screen)
    global model, mapimage, landcovermap, launching
    launching && display(screen, Figure().scene) # blank screen at launch
    println("Updating map")
    #FIXME I'm not sure date-1 is the real fix
    #FIXME repeatedly calling `visualisemap` causes the memory leak
    # (regardless of whether individuals are plotted or just the map)
    #mapimage[] = visualisemap(model, date[]-Day(1), landcovermap)
    mapimage[] = vismap(model, date[]-Day(1), landcovermap)
    display(screen, mapimage[])
end

function vismap(model::AgentBasedModel,date=nothing,landcover=nothing)
    # load and plot the map
    # Note: if the landcover map is supplied, it needs to be rotr90'ed
    isnothing(landcover) && (landcover = load(@param(world.landcovermap)))
    f = Figure()
    ax = Axis(f[1,1])
    hidedecorations!(ax)
    image!(f[1,1], landcover)
    ax.aspect = DataAspect()
    # check if there are individuals and plot them
    inds = model.datatables["individuals"]
    if iszero(size(inds)[1])
        @debug "No individual data to map"
        return f.scene
    end
    isnothing(date) && (date = inds.Date[end])
    for s in unique(inds.Species)
        points = @select!(@subset(inds, :Species .== s, :Date .== date),
                          :X, :Y)
        # The origin in Makie is in the bottom-left rather than in the top-left as
        # on the model map, so we have to invert the Y coordinates
        @transform!(points, :Y = size(model.landscape)[2] .- :Y)
        scatter!(f[1,1], Matrix{Float32}(points), markersize=8)
    end
    f.scene
end

function render_plot(screen)
    #FIXME causes segfault when called during a simulation?
    global model, launching
    launching && display(screen, Figure().scene) # blank screen at launch
    println("Updating plot")
    figure = populationtrends(model)
    display(screen, figure.scene)
end

on(delay) do d
    println("Delay is now $(round(d, digits=1)) seconds.") #XXX DEBUG
end

on(running) do r
    ticks = 0 #prevent overflows
    r ? println("Simulation started.") : println("Simulation stopped.") #XXX DEBUG
end

on(ticks) do t
    running[] && runsimulation()
end

datestring = () -> Dates.format(date[], "dd U yyyy")

function activateqmlfunctions()     
    @qmlfunction newsimulation
    @qmlfunction configwindow
    @qmlfunction loadsimulation
    @qmlfunction savesimulation
    @qmlfunction nextstep
    @qmlfunction previousstep
    @qmlfunction togglerunning
    @qmlfunction datestring
    @qmlfunction writeconfig
end

"""
    launch()

The main function that creates the application.
"""
function launch()
    global model, timer, launching
    launching = true

    activateqmlfunctions()
    timer = QTimer()

    GLMakie.activate!()
    ENV["QSG_RENDER_LOOP"] = "basic"
    QML.setGraphicsApi(QML.OpenGL)
    
    qmlfile = joinpath(dirname(@__FILE__), "main.qml")
    loadqml(qmlfile,
            timer = timer,
            vars = JuliaPropertyMap(guiproperties...,configproperties...),
            render_map_callback = @safe_cfunction(render_map, Cvoid, (Any,)),
            render_plot_callback = @safe_cfunction(render_plot, Cvoid, (Any,)))
    
    @emit showSplash()
    isnothing(model) && newsimulation()
    @emit closeSplash()
    launching = false
    @emit updateMakie()
    println("Launched Persefone Desktop.")
    exec()
end
